import json

from .config import get_env_reader


def get_cache_filename():
    return f".{get_env_reader()('ENV')}-cache.json"


def cache_read():
    try:
        with open(get_cache_filename(), 'r') as f:
            return json.load(f)
    except FileNotFoundError:
        return {}
    except json.JSONDecodeError:
        return {}


def cache_write(cache):
    with open(get_cache_filename(), 'w') as f:
        json.dump(cache, f, indent=2)


def cache_set(key):
    def with_value(value):
        cache = cache_read()
        cache[key] = value
        cache_write(cache)
        return value
    return with_value


def cache_get(key):
    cache = cache_read()
    return cache.get(key)


def retrieve_and_cache(key=None, arg=None, describe_func=None):

    def with_func(func):
        resp = cache_get(key)
        if resp is None:
            print(f"Cache miss key <{key}>")
            if arg is None:
                retrieve_result = func()
            else:
                retrieve_result = func(arg)
            cache_set(key)(retrieve_result)
        else:
            print(f"Cache hit key <{key}>")
        resp = cache_get(key)
        print(f"{key}: {describe_func(resp)}")
        return resp
    return with_func



import json
import requests
from python_graphql_client import GraphqlClient


def retrieve_graphql(url=None, query=None, variables=None, auth=None, log_prefix='Do a GraphQL query'):
    client = GraphqlClient(endpoint=url, auth=auth)
    prefix = f"{log_prefix} to {url}"
    print(prefix, 'variables', variables)
    response = client.execute(query, variables)
    print(prefix, 'response', response)
    return response


def retrieve_http(url=None, method=requests.get, log_prefix='Do a HTTP GET request', 
                  use_default_headers=True, auth=None, extra_headers={},
                  payload=None):
    prefix = f"{log_prefix} to {url}"

    def get_default_headers():
        headers = {}
        if use_default_headers:
            headers.update({
                'Content-Type': 'application/json'
            })
            if auth:
                headers['Authorization'] = auth
        headers.update(extra_headers)
        print(prefix, 'headers', headers)
        return headers

    print(prefix, 'payload', payload)
    response = method(
        url=url,
        data=None if payload is None else json.dumps(payload),
        headers=get_default_headers(),
    )
    print(prefix, 'response', response)
    data = response.json()
    print(prefix, 'JSON response', data)
    return data



from verifit.config import get_env_reader

get_env = get_env_reader()


def get_driver():
    return get_env('DRIVER')


def call_driver(drivers):
    def with_args(**kwargs):
        driver = get_driver()
        if driver in drivers.keys():
            return drivers.get(driver)(**kwargs)
    return with_args


def driver_is_one_of(drivers):
    return get_driver() in drivers



from datetime import datetime

from .cache import cache_set, cache_get
from .config import get_env_reader
from .date_tools import date_subtract_in_minutes
from .json_web_token import decode_token, get_token_expiration_date

get_env = get_env_reader()

ACCESS_TOKEN = 'accessToken'
EXPIRY_DATE = 'expiryDate'


def get_expiry_date(login_data):
    return datetime.fromisoformat(login_data.get(EXPIRY_DATE))


def login(driver=None, username=None, password=None):
    access_token = driver(username=username, password=password)
    decoded_token = decode_token(access_token)
    token_expiry_date = get_token_expiration_date(decoded_token)
    login_data = {
        ACCESS_TOKEN: access_token,
        EXPIRY_DATE: token_expiry_date.isoformat()
    }
    cache_set(username)(login_data)
    return login_data


def get_login_values_from_cache(username):
    user_cached_data = cache_get(username)
    if user_cached_data is None:
        print(f"Cache miss key <{username}>, no cached data")
        return None
    token_expiry_iso_string = user_cached_data.get('expiryDate', None)
    if token_expiry_iso_string is None:
        print(f"Cache miss key <{username}>, no expiration info")
        return None
    token_expiry_date = datetime.fromisoformat(token_expiry_iso_string)
    minutes = date_subtract_in_minutes(from_date=token_expiry_date, date_to_subtract=datetime.now())
    if minutes < 10:
        print(f"Cache miss key <{username}>, almost expired")
        return None
    print(f"Cache hit key <{username}>")
    login_data = {
        ACCESS_TOKEN: user_cached_data.get(ACCESS_TOKEN),
        EXPIRY_DATE: token_expiry_date
    }
    return login_data


def login_from_cache(driver=None, username=None, password=None):
    login_data = get_login_values_from_cache(username)
    if login_data is None:
        login_data = login(driver=driver, username=username, password=password)
    print(f"Caching login data <{login_data}>")
    return login_data




def date_subtract_in_minutes(from_date=None, date_to_subtract=None):
    difference = from_date - date_to_subtract
    minutes = difference.seconds // 60
    if difference.days < 0:
        minutes = minutes + difference.days * 24 * 60
    return minutes



import asyncio
import json

import websockets

from .config import get_env_reader

get_env = get_env_reader()

def ws_send_and_receive(server=None, ignore_list=None, data=None):

    wait_ms = int(get_env('WEBSOCKETS_WAIT_MS'))

    async def receive_async():
        async with websockets.connect(server) as websocket:
            print(f"Connected to server {server}")
            await websocket.send(json.dumps(data))
            print(f"Sent <{data}>")
            received = False
            while not received:
                received_data = await websocket.recv()
                if received_data in ignore_list:
                    print(f"Ignoring received message <{received_data}>")
                else:
                    received = True
        return received_data

    async def receive_async_with_timeout():
        seconds = wait_ms / 1000
        print(f"Sending and receiving for {seconds} sec from {server}")
        received_data = await asyncio.wait_for(
            receive_async(),
            seconds
        )
        print(f"Received <{received_data}>")
        return received_data

    def start_receiving():
        return asyncio.get_event_loop().run_until_complete(receive_async_with_timeout())


    return json.loads(start_receiving())



from functools import cache
from multiprocessing import Process
import time
import queue
from flask import Flask, request, json


@cache
def api():
    return Flask('webhook-observer-server')


@cache
def port():
    return 5768


@cache
def endpoint():
    return '/webhook'


@cache
def webhook_url():
    return f"http://localhost:{port()}{endpoint()}"


KEY_QUEUE = 'queue'
KEY_CALLS_COUNT = 'calls_count'
KEY_STATUS_TO_RETURN = 'status_to_return'
KEY_SUCCEED_AT_ATTEMPT_NO = 'succeed_at_attempt_no'


@cache
def get_store():
    return {}


@api().route(endpoint(), methods=['POST'])
def webhook_observer():
    store = get_store()
    q = store.get(KEY_QUEUE)
    webhook_response = request.json
    print(f"Payload: {webhook_response}")
    status_to_return = store.get(KEY_STATUS_TO_RETURN)
    succeed_at_attempt_no = store.get(KEY_SUCCEED_AT_ATTEMPT_NO)
    calls_count = store.get(KEY_CALLS_COUNT)
    calls_count = calls_count + 1
    store[KEY_CALLS_COUNT] = calls_count
    if succeed_at_attempt_no > 0:
        if calls_count >= succeed_at_attempt_no:
            status_to_return = 202
    success = 200 <= status_to_return <= 299
    local_response = {"success": success}
    q.put(webhook_response if success else local_response)
    return json.dumps(local_response), status_to_return


def start_webhook_observer(q, s, a):
    store = get_store()
    store[KEY_QUEUE] = q
    store[KEY_CALLS_COUNT] = 0
    store[KEY_STATUS_TO_RETURN] = s
    store[KEY_SUCCEED_AT_ATTEMPT_NO] = a
    api().run(port=port())


def webhook_server_start(*, queue, status=200, succeed_at_attempt_no=0):
    server = Process(target=start_webhook_observer, args=(queue, status, succeed_at_attempt_no))
    server.start()
    time.sleep(2)
    return server


def queue_is_empty(q):
    try:
        q.get(timeout=1)
        return False
    except queue.Empty:
        return True
    except:
        assert False



import os
from functools import cache

from dotenv import dotenv_values


@cache
def config():
    environment_name = os.environ.get('ENV', 'dev')
    os.environ['ENV'] = environment_name
    values = {
        **dotenv_values(f".{environment_name}.env"),
        **os.environ
    }
    return values


def get_env_reader():
    store = config()

    def get_env(key):
        return store.get(key)

    return get_env



import datetime
import jwt


def get_token_expiration_date(decoded_token):
    token_expiry_ms = decoded_token.get('exp', None)
    assert token_expiry_ms is not None
    token_expiry_date = datetime.datetime.fromtimestamp(token_expiry_ms)
    return token_expiry_date


def decode_token(access_token):
    decoded_token = jwt.decode(access_token, algorithms=["RS256", "HS256"], options={"verify_signature": False})
    return decoded_token
